@@ -49,7 +49,9 @@ known_restrictions = frozenset({
     'needs-reboot',
     'needs-recommends',
     'needs-root',
+    'needs-sudo',
     'rw-build-tree',
+    'skip-foreign-architecture',
     'skip-not-installable',
     'skippable',
     'superficial',
@@ -61,7 +63,7 @@ known_restrictions = frozenset({
 # - str: a capability that is required by the restriction
 # - set of two or more str: at least one of these caps is required
 #   by the restriction
-RESTRICTIONS_REQUIRE_CAPS = {
+RESTRICTIONS_REQUIRE_CAPS: Dict[str, FrozenSet[Union[str, FrozenSet[str]]]] = {
     'breaks-testbed': frozenset({'revert-full-system'}),
     'needs-internet': frozenset({'has_internet'}),
     'needs-reboot': frozenset({'reboot'}),
@@ -70,7 +72,14 @@ RESTRICTIONS_REQUIRE_CAPS = {
         frozenset({'isolation-container', 'isolation-machine'}),
     }),
     'isolation-machine': frozenset({'isolation-machine'}),
-}       # type: Dict[str, FrozenSet[Union[str, FrozenSet[str]]]]
+}
+
+
+# Keys are restrictions
+# Values are sets of dependency strings
+RESTRICTIONS_IMPLY_DEPS = {
+    'needs-sudo': frozenset({'sudo'}),
+}
 
 
 class Unsupported(Exception):
@@ -140,8 +149,14 @@ class Test:
         self.command = command
         self.restrictions = set(restrictions)
         self.features = set(features)
-        self.depends = depends
+        self.depends = list(depends)
         self.synth_depends = synth_depends
+
+        for r in self.restrictions:
+            for d in RESTRICTIONS_IMPLY_DEPS.get(r, frozenset()):
+                if d not in self.depends:
+                    self.depends.append(d)
+
         # None while test hasn't run yet; True: pass, False: fail
         self.result = None
         self.skipped = False
@@ -269,7 +284,8 @@ def _debian_check_unknown_fields(name, record):
         raise Unsupported(name, 'unknown field %s' % unknown_keys.pop())
 
 
-def _debian_packages_from_source(srcdir):
+def _debian_packages_from_source(srcdir, testbed_arch, test_arch):
+    test_arch_is_foreign = (test_arch != testbed_arch)
     packages = []
     packages_no_arch = []
 
@@ -282,11 +298,17 @@ def _debian_packages_from_source(srcdir):
                 st.get('Package-type', 'deb') != 'deb':
             continue
         arch = st['Architecture']
-        if arch in ('all', 'any'):
-            packages.append(st['Package'])
+        binary_package = st['Package']
+        packages_no_arch.append(binary_package)
+        if arch == 'all':
+            packages.append(binary_package)
         else:
-            packages.append('%s [%s]' % (st['Package'], arch))
-        packages_no_arch.append(st['Package'])
+            if test_arch_is_foreign:
+                binary_package += ':' + test_arch
+            if arch == 'any':
+                packages.append(binary_package)
+            else:
+                packages.append('%s [%s]' % (binary_package, arch))
 
     return (packages, packages_no_arch)
 
@@ -301,30 +323,17 @@ def _filter_variables(deps):
     return ', '.join(deplist)
 
 
-def _debian_recommends_from_source(srcdir, testbed_arch):
+def _debian_recommends_from_source(srcdir):
     deps = ''
     for st in parse_rfc822(os.path.join(srcdir, 'debian/control')):
         if 'Recommends' in st:
-            deps += st['Recommends'] + ','
+            deps += st['Recommends']
 
     # Because we parse the source d/control file instead of the binary
     # control file, we may run into variables. See bug 1008206 for
     # some discussion.
     deps = _filter_variables(deps)
 
-    # resolve arch specific dependencies and build profiles
-    perl = subprocess.Popen(['perl', '-'], stdin=subprocess.PIPE,
-                            stdout=subprocess.PIPE)
-    code = '''use Dpkg::Deps;
-              $dep = deps_parse('%s', reduce_restrictions => 1,
-                                host_arch => '%s');
-              $out = $dep->output();
-              print $out, "\\n";
-              ''' % (deps, testbed_arch)
-    deps = perl.communicate(code.encode('UTF-8'))[0].decode('UTF-8').strip()
-    if perl.returncode != 0:
-        raise InvalidControl('source', 'Invalid Recommends')
-
     deps = [d.strip() for d in deps.split(',')]
 
     # The following check is a bit silly as there were apparently no
@@ -336,32 +345,15 @@ def _debian_recommends_from_source(srcdir, testbed_arch):
     return deps
 
 
-def _debian_build_deps_from_source(srcdir, testbed_arch):
+def _debian_build_deps_from_source(srcdir):
     deps = ''
     for st in parse_rfc822(os.path.join(srcdir, 'debian/control')):
         if 'Build-depends' in st:
-            deps += st['Build-depends']
+            deps += st['Build-depends'].rstrip(',')
         if 'Build-depends-indep' in st:
-            deps += ', ' + st['Build-depends-indep']
+            deps += ', ' + st['Build-depends-indep'].rstrip(',')
         if 'Build-depends-arch' in st:
-            deps += ', ' + st['Build-depends-arch']
-
-    # resolve arch specific dependencies and build profiles
-    perl = subprocess.Popen(['perl', '-'], stdin=subprocess.PIPE,
-                            stdout=subprocess.PIPE)
-    code = '''use Dpkg::Deps;
-              $supports_profiles = ($Dpkg::Deps::VERSION gt '1.04');
-              $dep = deps_parse('%s', reduce_arch => 1,
-                                reduce_profiles => $supports_profiles,
-                                build_dep => 1, host_arch => '%s');
-              $out = $dep->output();
-              # fall back to ignoring build profiles
-              $out =~ s/ <[^ >]+>//g if (!$supports_profiles);
-              print $out, "\\n";
-              ''' % (deps, testbed_arch)
-    deps = perl.communicate(code.encode('UTF-8'))[0].decode('UTF-8').strip()
-    if perl.returncode != 0:
-        raise InvalidControl('source', 'Invalid build dependencies')
+            deps += ', ' + st['Build-depends-arch'].rstrip(',')
 
     deps = [d.strip() for d in deps.split(',')]
 
@@ -398,7 +390,7 @@ def _debian_check_dep(testname, dep):
     return (m.group('package'), m.group('version'))
 
 
-def _synthesize_deps(dep, testbed_arch):
+def _synthesize_deps(dep, test_arch):
     '''Ensure that apt can install synthesized Depends
 
     Test Depends may have architecture qualifiers we need to check,
@@ -417,7 +409,7 @@ def _synthesize_deps(dep, testbed_arch):
                                 stdout=subprocess.PIPE)
         code = '''use Dpkg::Deps;
                   print deps_parse('foo [%s]')->arch_is_concerned('%s') ? 'y' : 'n';
-                  ''' % (m.group('arch'), testbed_arch)
+                  ''' % (m.group('arch'), test_arch)
         arch_matches = perl.communicate(code.encode('UTF-8'))[0].decode('UTF-8').strip()
         if perl.returncode != 0:
             raise InvalidControl('source', 'Invalid (Test-)Depends architecture qualifiers')
@@ -428,7 +420,7 @@ def _synthesize_deps(dep, testbed_arch):
         return None
 
 
-def _parse_debian_depends(testname, dep_str, srcdir, testbed_arch):
+def _parse_debian_depends(testname, dep_str, srcdir, testbed_arch, test_arch):
     '''Parse Depends: line in a Debian package
 
     Split dependencies (comma separated), validate their syntax, and expand @,
@@ -437,9 +429,15 @@ def _parse_debian_depends(testname, dep_str, srcdir, testbed_arch):
     This may raise an InvalidControl exception if there are invalid
     dependencies.
     '''
+    test_arch_is_foreign = (test_arch != testbed_arch)
     deps = []
     synthdeps = []
-    (my_packages, my_packages_no_arch) = _debian_packages_from_source(srcdir)
+    (my_packages, my_packages_no_arch) = _debian_packages_from_source(
+        srcdir,
+        testbed_arch,
+        test_arch,
+    )
+
     for alt_group_str in dep_str.split(','):
         alt_group_str = alt_group_str.strip()
         if not alt_group_str:
@@ -450,15 +448,15 @@ def _parse_debian_depends(testname, dep_str, srcdir, testbed_arch):
             for d in my_packages:
                 adtlog.debug('synthesised dependency %s' % d)
                 deps.append(d)
-                s = _synthesize_deps(d, testbed_arch)
+                s = _synthesize_deps(d, test_arch)
                 if s:
                     synthdeps.append(s)
         elif alt_group_str == '@builddeps@':
-            for d in _debian_build_deps_from_source(srcdir, testbed_arch):
+            for d in _debian_build_deps_from_source(srcdir):
                 adtlog.debug('synthesised dependency %s' % d)
                 deps.append(d)
         elif alt_group_str == '@recommends@':
-            for d in _debian_recommends_from_source(srcdir, testbed_arch):
+            for d in _debian_recommends_from_source(srcdir):
                 adtlog.debug('synthesised dependency %s' % d)
                 deps.append(d)
         else:
@@ -468,8 +466,10 @@ def _parse_debian_depends(testname, dep_str, srcdir, testbed_arch):
                 if pkg not in my_packages_no_arch:
                     synthdep_alternatives = []
                     break
-                s = _synthesize_deps(dep, testbed_arch)
+                s = _synthesize_deps(dep, test_arch)
                 if s:
+                    if test_arch_is_foreign:
+                        s += ':' + test_arch
                     synthdep_alternatives.append(s)
             if synthdep_alternatives:
                 adtlog.debug('marked alternatives %s as a synthesised dependency' % synthdep_alternatives)
@@ -477,8 +477,28 @@ def _parse_debian_depends(testname, dep_str, srcdir, testbed_arch):
                     synthdeps.append(synthdep_alternatives)
                 else:
                     synthdeps.append(synthdep_alternatives[0])
+            if test_arch_is_foreign:
+                for mine in my_packages:
+                    # ignore [arch] filters for matching packages
+                    if alt_group_str + ':' + test_arch == mine.split(' ')[0]:
+                        adtlog.debug('%s is from our source package, adding arch qualifier for cross-testing' % s)
+                        alt_group_str += ':' + test_arch
+                        break
             deps.append(alt_group_str)
 
+    if test_arch_is_foreign and "build-essential" in deps:
+        # Ensure we are installing the testbed-native version of
+        # build-essential by specifying :native. This is needed because the
+        # build-essential package is not Multi-Arch: foreign, so we need to
+        # make sure we are always picking up the package for the testbed
+        # architecture, not for the foreign test architecture.
+        deps.remove('build-essential')
+        deps.append('build-essential:native')
+        # also add build-dependencies on libc-dev and libstdc++-dev,
+        # as a workaround to #815172.
+        deps.append('libc-dev:%s' % test_arch)
+        deps.append('libstdc++-dev:%s' % test_arch)
+
     return (deps, synthdeps)
 
 
@@ -521,8 +541,8 @@ def _matches_architecture(host_arch, arch_wildcard):
     return True
 
 
-def _check_architecture(name, testbed_arch, architectures):
-    '''Check if testbed_arch is supported by the architectures
+def _check_architecture(name, test_arch, architectures):
+    '''Check if test_arch is supported by the architectures
 
     The architecture list comes in two variants, positive: only this
     arch is supported (arch may be a wildcard) and negative: this arch
@@ -545,9 +565,9 @@ def _check_architecture(name, testbed_arch, architectures):
     for arch in architectures:
         if arch[0] == "!":
             any_negative = True
-            if _matches_architecture(testbed_arch, arch[1:]):
+            if _matches_architecture(test_arch, arch[1:]):
                 raise Unsupported(name, "Test declares architecture as not " +
-                                  "supported: %s" % testbed_arch)
+                                  "supported: %s" % test_arch)
         if arch[0] != "!":
             any_positive = True
 
@@ -558,18 +578,18 @@ def _check_architecture(name, testbed_arch, architectures):
                               "others aren't")
         arch_matched = False
         for arch in architectures:
-            if _matches_architecture(testbed_arch, arch):
+            if _matches_architecture(test_arch, arch):
                 arch_matched = True
 
         if not arch_matched:
             raise Unsupported(name, "Test lists explicitly supported " +
                               "architectures, but the current architecture " +
-                              "%s isn't listed." % testbed_arch)
+                              "%s isn't listed." % test_arch)
 
 
-def parse_debian_source(srcdir, testbed_caps, testbed_arch, control_path=None,
-                        auto_control=True, ignore_restrictions=(),
-                        only_tests=()):
+def parse_debian_source(srcdir, testbed_caps, testbed_arch, test_arch,
+                        control_path=None, auto_control=True,
+                        ignore_restrictions=(), only_tests=()):
     '''Parse test descriptions from a Debian DEP-8 source dir
 
     @ignore_restrictions: If we would skip the test due to these restrictions,
@@ -656,7 +676,9 @@ def parse_debian_source(srcdir, testbed_caps, testbed_arch, control_path=None,
                     test_names[0],
                     record.get('Depends', '@'),
                     srcdir,
-                    testbed_arch)
+                    testbed_arch,
+                    test_arch,
+                )
                 if 'Test-command' in record:
                     raise InvalidControl('*', 'Only one of "Tests" or '
                                          '"Test-Command" may be given')
@@ -668,7 +690,7 @@ def parse_debian_source(srcdir, testbed_caps, testbed_arch, control_path=None,
                 for n in test_names:
                     try:
                         _debian_check_unknown_fields(n, record)
-                        _check_architecture(n, testbed_arch, architectures)
+                        _check_architecture(n, test_arch, architectures)
 
                         test = Test(n, os.path.join(test_dir, n), None,
                                     restrictions, features, depends, synth_depends)
@@ -685,14 +707,16 @@ def parse_debian_source(srcdir, testbed_caps, testbed_arch, control_path=None,
                     command,
                     record.get('Depends', '@'),
                     srcdir,
-                    testbed_arch)
+                    testbed_arch,
+                    test_arch,
+                )
                 if feature_test_name is None:
                     command_counter += 1
                     name = 'command%i' % command_counter
                 else:
                     name = feature_test_name
                 _debian_check_unknown_fields(name, record)
-                _check_architecture(name, testbed_arch, architectures)
+                _check_architecture(name, test_arch, architectures)
                 test = Test(name, None, command, restrictions, features,
                             depends, synth_depends)
                 test.check_testbed_compat(testbed_caps, ignore_restrictions)
